/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <ctype.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "support/getopt.h"
#include "support/getopt.c"
#include "support/getopt1.c"
#endif

#include "jproxygw.h"

void print_version(struct proxyinfo* pi, const char* proxy_name);
void print_help(struct proxyinfo* pi, const char* proxy_name);

extern struct loginfo_st loginfo;
int timeout;
int should_read_config;
int should_terminate;
int chlds_exited;

static int install_signal_handlers(void);

static struct option const long_option_arr[] =
{
	{ "configfile", required_argument, NULL, 'f' },
	{ "encrypt",    no_argument,       NULL, 'e' },
	{ "help",       no_argument,       NULL, 'h' },
	{ "inetd",      no_argument,       NULL, 'i' },
	{ "single",     no_argument,       NULL, 's' },
#ifdef HAVE_LIBWRAP
	{ "tcpwrap",    no_argument,       NULL, 'w' },
#endif
	{ "test",       no_argument,       NULL, 't' },
	{ "version",    no_argument,       NULL, 'V' },
	{ "version",    no_argument,       NULL, 'v' },
	{ NULL,         0,                 NULL,  0  }
};

int jproxygw_main(int argc, char** argv,
		  struct protocolhandler proto_handler) {

	/* We bind to 0.0.0.0, but our real IP is the one returned by the
	 * PASV command */

	struct proxyinfo pi;
	struct serverinfo si;
	int ret;
	const char* option;

#ifdef RLIMIT_CORE
	/* set RLIMIT_CORE to 0 */
	struct rlimit rl = { 0, 0 };

	if (setrlimit(RLIMIT_CORE, &rl) < 0) {
		perror("setrlimit");
		return -1;
	}
#endif

	pi.si = &si;
	pi.ci = (struct clientinfo*) 0;
	pi.protocol = &proto_handler;

	/* set the name of the config file to an initial value, it may be
	 * overwritten later
	 * */
	si.conffilename = strdup(proto_handler.defaultconfigfile);
	enough_mem(si.conffilename);

	/* default: Multithread, i.e. fork for each connection */
	si.multithread = 1;
	si.tcp_wrapper = 0;
	si.servertype = SERVERTYPE_STANDALONE;
	si.main_server_pid = 0;
	si.chrooted = 0;
	si.hostcache = (struct hostent_list*) 0;
	si.connection.server_name = (char*) 0;
	si.connection.client_name = (char*) 0;
	si.connection.listen_specs = (char*) 0;
	FD_ZERO(&si.connection.listen_set.set);
	si.connection.listen_set.maxfd = -1;

	if (strrchr(argv[0], '/')) {
		si.binaryname = strdup(strrchr(argv[0], '/') + 1);
	} else {
		si.binaryname = strdup(argv[0]);
	}
	enough_mem(si.binaryname);

	/* parse the command line */
	while ((ret=getopt_long(argc, argv, "isewtvVhf:", long_option_arr, NULL)) != EOF) {
		switch (ret) {
			case 0: break;
			case 'i':
				si.servertype = SERVERTYPE_INETD;
				si.multithread = 0;
				break;
			case 's':
				si.multithread = 0;
				break;
#ifdef HAVE_LIBWRAP
			case 'w':
				si.tcp_wrapper = 1;
				break;
#endif
			case 't':
				should_terminate = 1;
				break;
			case 'e':
				encrypt_password();
				exit(0);
			case 'v':
			case 'V':
				print_version(&pi, proto_handler.proxy_name);
				exit(0);
			case 'h':
				print_help(&pi, proto_handler.proxy_name);
				exit(0);
			case 'f':
				if (set_conffilename(optarg, &pi) < 0) {
					jlog(2, "something is wrong with the name of the configuration file");
					return 1;
				}
				break;
			default:
				break;
		}
	}

	si.connection.boundsocket_list = (int*) 0;
	si.connection.connected_clients = (struct connliststruct*) 0;

	memset(&loginfo, 0, sizeof(struct loginfo_st));
	loginfo.debuglevel = 6;
	loginfo.stderr_open = 1;
	si.ready_to_serve = SVR_LAUNCH_CMDLINE;
	si.pidfile_created = 0;
	si.runasuser.username = (char*) 0;
	si.runasuser.groupname = (char*) 0;

	si.connection.serversocket = -1;
	si.connection.transparent = TRANSPARENT_NO;
	si.connection.server_ip
		= si.connection.client_ip
		= si.connection.addr_to_server
		= si.connection.addr_to_client
		= (unsigned long int) UINT_MAX;
	si.throughput = -1.0;
	si.throughput_send = -1.0;
	si.throughput_receive = -1.0;

	/* init other parts */
	event_init();
	filter_init();

	/* initialize the modules, they will hook in a function that gets
	 * called a few lines down by event_fire() that will add their
	 * configuration data */
	if (proto_handler.proto_init_modules() < 0) {
		return 1;
	}

	/* set hooks */
	config_register_hook(0, proto_handler.proto_config_install);

	/* XXX check return value */
	event_fire(EVENT_PROXY_INIT_CONFIG, &pi);

	if (proto_handler.proto_read_config_before(&pi) < 0) {
		return -1;
	}
	/* Read the configuration. We do not apply strip_chrooted_path because
	 * this is the first time we read the configuration file. So we don't
	 * even know the directory of the chroot */
	ret = read_config(si.conffilename, &pi);
	if (ret) {
		return -1;
	}
	if (proto_handler.proto_read_config_after(&pi) < 0) {
		return -1;
	}

	/* a check... */
	if ((option = config_get_option("changeroot"))
			&& strcmp(option, "never") != 0) {
		/* we should change the root directory at some time in the
		 * future */
		if (getuid() != 0) {
			jlog(5, "Changeroot requested but not started as superuser but with uid: %d", getuid());
			return -1;
		}
	}

	/* Drop privileges right after the start of the program. Right after
	 * reading the configuration file */
	if (stage_action("start", &pi) < 0) {
		return -1;
	}

	/* init the logfiles */
	if (changeid(PRIV, UID, "log_init()", &pi) < 0) { return -1; }
	if (log_init(&pi) < 0) {
		changeid(UNPRIV, EUID, "log_init()", &pi);
		return -1;
	}
	if (changeid(UNPRIV, EUID, "log_init()", &pi) < 0) { return -1; }

	si.ready_to_serve = SVR_LAUNCH_LOGFILES;

	install_signal_handlers();

	/* XXX check return value */
	event_fire(EVENT_PROXY_INIT_COMPLETE, &pi);

	/* okay, wait for a client to connect and call the appropriate
	 * handler if this has happened */
	if (!should_terminate && si.servertype == SERVERTYPE_STANDALONE) {
		ret = waitclient(&pi, &proto_handler); 
		/* An error occured or we should terminate */
	} else if (!should_terminate && si.servertype == SERVERTYPE_INETD) {
		close(1);
		dup2(0, 10);
		close(0);
		if (inetd_connected(10, &pi, &proto_handler) < 0) {
			return -1;
		}
	}
	jlog(8, "Exiting...");

	/* deallocate the protocol */
	if (proto_handler.proto_deallocate(&pi) < 0) {
		jlog(5, "Error deallocating the protocol");
	}

	/* deinit the parts we've inited before */
	event_deinit();
	filter_deinit();
	removepidfile(&pi);

	/* deinit the logfiles */
	if (changeid(PRIV, UID, "log_deinit()", &pi) < 0) { return -1; }
	log_deinit(&pi);
	if (changeid(UNPRIV, EUID, "log_deinit()", &pi) < 0) { return -1; }

	/* this routine will delete the configuration as well so it should be
	 * the last one that is called */
	closedescriptors(&pi);
	if (should_terminate) {
		/* correct shutdown */
		return 0;
	} else {
		/* an error occurred */
		return 1;
	}
}


int daemonize(void) {
	int childpid;

	if( (childpid = fork()) < 0) {
		return -1;
	} else if (childpid > 0) {
		exit(0);
	}

	errno = 0;
	chdir("/");
	setsid();

	/* XXX enable this for per-client malloc debugging */
	/* dmalloc_debug_setup("debug=0x4f40d03,log=/tmp/logfile"); */
	return 0;
}


void read_default_conf(int signo) {
	should_read_config = 1;
}

int reread_config(struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;
	const char* filename;
	struct protocolhandler* proto_handler = pi->protocol;

	int ret;
	sigset_t sigset, oldset;
	jlog(6, "SIGHUP received - Rereading config file");

	sigemptyset(&sigset);
	sigemptyset(&oldset);
	/* block other SIGHUP signals - actually this is not really necessary
	 * since our signal handler just sets a variable... */
	sigaddset(&sigset, SIGHUP);
	/* we block this one because if we re-create the configuration, we also
	 * recreate the counters for the limit option. However, imagine we are
	 * right in the middle of re-creating the tag list and a client exits.
	 * Then we try to unregister it and we would try to read the counters
	 * that have not yet been recreated from the list of connected clients.
	 * */
	sigaddset(&sigset, SIGCHLD);
	while ((ret = sigprocmask(SIG_BLOCK, &sigset,
		&oldset)) < 0 && errno == EINTR)  {}
	if (ret < 0) {
		jlog(2, "Error blocking signals: %s", strerror(errno));
	}
	/* see if we have to have a special chrooted filename as long as we
	 * have not yet deleted the configuration. Note that filename points
	 * into pi->si->conffilename, so there is no need to free this pointer
	 * */
	filename = strip_chrooted_path(pi->si->conffilename);
	if (proto_handler->proto_read_config_before(pi) < 0) {
		return -1;
	}
	config_delete_config();
	ret = read_config(filename, pi);
	if (ret) {
		/* the following line is a BADF if we had a SIGHUP */
		/* XXX or should we silently fall back to some backup
		 * configuration? */
		jlog(1, "Error rereading config file. Exiting.");
		exit(2);
	}
	/* re-register all connected clients */
	config_counter_add_connected(si->connection.connected_clients,
				     si->servertype, &(si->hostcache));
	/* re-sort filters */
	filters_sort();

	/* I don't know exactly if this should be here or above or below... */
	if (proto_handler->proto_read_config_after(pi) < 0) {
		return -1;
	}

	if (log_reinit(pi) < 0) {
		return -1;
	}

	if (strcmp(config_get_option("listen"), si->connection.listen_specs)) {
		/* if we have a difference, we get here */
		/* XXX create event function ? */
		unlisten_on_ifaces(pi);
		if (listen_on_ifaces(pi) < 0) {
			jlog(1, "Could not bind to all requested interfaces "
				"after re-reading the config file. Exiting...");
			return -1;
		}
	} else {
		jlog(8, "listen parameter doesn't seem to have changed.");
	}

	/* XXX check return type */
	event_fire(EVENT_PROXY_REREAD_CONFIG, pi);

	should_read_config = 0;
	while ((ret = sigprocmask(SIG_UNBLOCK, &sigset,
		&oldset)) < 0 && errno == EINTR) {}
	if (ret < 0) {
		jlog(2, "Error unblocking signal mask: %s", strerror(errno));
		return -1;
	}
	return 0;
}

void terminate (int signo) {
	/* exit is not POSIX-reentrant but in SVR4 SVID */
	/* exit(0); */
	should_terminate = 1;
}


int stage_action(const char* stage, struct proxyinfo* pi) {
	char* reason = char_append("stage_action() in stage ", stage);

	changeid(PRIV, UID, reason, pi);
	free(reason);
	if (change_root(stage, pi) < 0) {
		changeid(UNPRIV, EUID, "change_root() failed", pi);
		return -1;
	}
	if (dropprivileges(stage, pi) < 0) {
		changeid(UNPRIV, EUID, "dropprivileges() failed", pi);
		return -1;
	}
	changeid(UNPRIV, EUID, "log_init()", pi);
	return 0;
}


void print_version(struct proxyinfo* pi, const char* proxy_name) {
	printf("%s", proxy_name);
#ifdef HAVE_CRYPT
	printf("  -  crypt support enabled");
#else
	printf("  -  without crypt support");
#endif


#ifdef HAVE_LINUX_NETFILTER_IPV4_H
	printf("  -  netfilter support enabled");
#else
	printf("  -  without netfilter support");
#endif
	printf("\n");

#ifdef HAVE_LIBWRAP
	/* XXX it would be cool to display the libwrap service names here
	 * but then we must read the config file before... */
	printf("libwrap support enabled");
#else
	printf("without libwrap support");
#endif

#ifdef HAVE_SIOCGIFADDR
	printf(" - can get IPs from interfaces");
#else
	printf(" - can't get IPs from interfaces");
#endif

#ifdef HAVE_ICMP_SUPPORT
	printf(" - ICMP support");
#else
	printf(" - no ICMP support");
#endif
	printf("\n");

	printf("Using configuration file: %s\n", pi->si->conffilename);
}

void print_help(struct proxyinfo* pi, const char* proxy_name) {
	print_version(pi, proxy_name);
	printf("usage: jftpgw [OPTION]\n\n");
	printf("Valid options:\n");
	printf("  -h, --help                Display this help text\n");
	printf("  -e, --encrypt             Use jftpgw to obtain an encrypted password\n");
	printf("  -f, --configfile file     Load file instead of default config file\n");
	printf("  -i, --inetd               Run jftpgw from inetd superserver\n");
	printf("  -s, --single              Run jftpgw single threaded (do not fork)\n");
	printf("  -t, --test                Start jftpgw, test configuration and terminate\n");
#ifdef HAVE_LIBWRAP
	printf("  -w, --tcpwrap             Use libwrap for access control\n");
#endif
	printf("  -V, -v, --version         Display the version\n");
	printf("\nReport bugs to Joachim Wieland <joe@mcknight.de>\n");
}


int removepidfile(struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;
	int i = 0;

	const char* option;
	if (getpid() != si->main_server_pid) {
		/* the program has not become a daemon */
		return 0;
	}
	if (!si->pidfile_created) {
		/* we have not yet created a pidfile */
		return 0;
	}

	option = config_get_option("pidfile");
	if (option) {
		if (changeid(PRIV, UID,
			"Changing ID to root (unlink pidfile)", pi) < 0) {
			return -1;
		}
		i = unlink(option);
		if (i) {
			jlog(3, "Could not unlink the pidfile %s: %s",
					option, strerror(errno));
		}
		if (changeid(UNPRIV, EUID,
			"Changing id back (deleting pidfile)", pi) < 0) {
			return -1;
		}
	}
	return i;
}

void sayterminating(void) {
	/* XXX name! */
	jlog(6, "jftpgw terminating");
}

void closedescriptors(struct proxyinfo* pi) {
	struct serverinfo* si = pi->si;

	int i;
	if (si->conffilename) {
		free(si->conffilename);
	}

	config_delete_config();
	config_delete_backup();
	config_deinit();

	if (si->runasuser.username) {
		free(si->runasuser.username);
		si->runasuser.username = 0;
	}
	if (si->runasuser.groupname) {
		free(si->runasuser.groupname);
		si->runasuser.groupname = 0;
	}
	free_errstr();

	free(si->binaryname);

	if (si->connection.boundsocket_list) {
		for (i = 0; i < si->connection.boundsocket_niface; i++) {
			close(si->connection.boundsocket_list[i]);
		}
		free(si->connection.boundsocket_list);
	}

	if (si->connection.server_name) {
		free(si->connection.server_name);
	}
	if (si->connection.client_name) {
		free(si->connection.client_name);
	}
	if (si->connection.listen_specs) {
		free(si->connection.listen_specs);
	}

	hostent_destroy(si->hostcache);
	si->hostcache = (struct hostent_list*) 0;
}


/* This function gets a path or a file name and checks if the path we chrooted
 * into is part of this path/file name. If it is and if we are chrooted, it
 * strips of the path where we chrooted into and returns the part that is
 * relative to the chrooted path.
 *
 * example:
 *
 * path (argument): /var/jftpgw/chrootdir/conf/jftpgw.conf
 * chrootdir:       /var/jftpgw/chrootdir
 *
 * returns:         /conf/jftpgw.conf
 */

const char* strip_chrooted_path(const char* fname) {
	struct stat st;

	const char* start;
	const char* chrootdir = config_get_option("changerootdir");

	if (!chrootdir) {
		/* return the same path */
		return fname;
	}
	if (stat(fname, &st) == 0) {
		/* the file or directory seems to exist */
		return fname;
	}
	/* okay, maybe we are chrooted, see if the logfile path lies in the
	 * chrooted dir */
	if (fname && chrootdir) {
		/* chrootdir is shorter than fname */
		if (strncmp(fname, chrootdir, strlen(chrootdir)) == 0) {
			/* Yeah, we're inside the chrooted dir, strip it
			 * off.
			 * Calculate where to split, make sure, the new path
			 * starts with a slash. */
			start = fname + strlen(chrootdir);
			while (start > fname && *start != '/') {
				start --;
			}
			jlog(8, "stripped the chrootdir %s from filename %s: using %s now", chrootdir, fname, start);
			return start;
		}
	}
	/* maybe we are not chrooted or the chrootdir is not part of the
	 * path...*/
	return fname;
}

static
int install_signal_handlers(void) {
	struct sigaction sa, cf;

	/* Install the initial signal handlers */

	/* for SIGCHLD install just the reap function
	 *
	 * the register/unregister thing is installed after we've bound
	 * successfully */

	sa.sa_handler = reap_chld_info;
	sigemptyset (&sa.sa_mask);
#ifndef WIN32
	sa.sa_flags = SA_RESTART;
#endif
	sigaction (SIGCHLD, &sa, 0);

	cf.sa_handler = read_default_conf;
	should_read_config = 0;
	sigemptyset(&cf.sa_mask);
#ifndef WIN32
	cf.sa_flags = SA_RESTART;
#endif
	sigaction (SIGHUP, &cf, 0);

	cf.sa_handler = terminate;
	sigemptyset(&cf.sa_mask);
	sigaction (SIGINT, &cf, 0);
	sigaction (SIGTERM, &cf, 0);
	sigaction (SIGQUIT, &cf, 0);
	sigaction (SIGABRT, &cf, 0);

	/* Ignore SIGALRM */
	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	sigaction(SIGALRM, &sa, 0);

	return 0;
}
