/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <syslog.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "jproxygw.h"
#include "log.h"

#define LOG_MAX_LINE_LENGTH 768

#ifndef HAVE_SNPRINTF
#   include "snprintf.c"
#else
#   ifndef HAVE_VSNPRINTF
#     include "snprintf.c"
#   endif
#endif

struct loginfo_st loginfo;
static int get_log_priority(const char* log_priority_opt);

/*
 * This routine logs messages to either the log file or the syslog function.
 */
void jlog(int debuglevel, const char *fmt, ...) {
	va_list args;
	time_t now;
	FILE *cf;

	static char now_str[LOG_STRFTIME_LENGTH];
	static char str[LOG_MAX_LINE_LENGTH];

	if (debuglevel > loginfo.debuglevel) {
		return;
	}

	/* we may dump the log to stderr if we have not yet
	 * opened a log file or the syslog */
	/* but don't log if we are run by inetd */
	if (!loginfo.logf
		&& !loginfo.syslog_debug
		&& loginfo.servertype == SERVERTYPE_INETD
		&& debuglevel > 5) {

		return;
	}

	va_start(args, fmt);
	if (!loginfo.syslog_debug || !loginfo.syslog_opened) {
		/* log via files */
		now = time(NULL);
		/* we log in the following format: Oct 07 17:41:59,
		 * i.e. MMM DD HH:MM:SS */
		strftime(now_str, LOG_STRFTIME_LENGTH,
				"%b %d %H:%M:%S", localtime(&now));
		if (!(cf = loginfo.logf)) {
			if (loginfo.stderr_open) {
				cf = stderr;
			} else {
				/* stderr is closed as well... */
				return;
			}
		}

		fprintf(cf, "%s [%ld]: ", now_str, (long int) getpid());
		vsnprintf(str, LOG_MAX_LINE_LENGTH - 1, fmt, args);
		/* kill linefeeds */
		replace_not_larger(str, "\r", "");
		/* kill newlines at the end */
		while (strlen(str) && (str[strlen(str) - 1] == '\n')) {
			str[strlen(str) - 1] = '\0';
		}
		/* replace \n with \\n within the string */
		if (strchr(str, '\n')) {
			/* we have at least one newline character */
			char* new_str = replace_larger(str, "\n", "\\n");
			fprintf(cf, "%s\n", new_str);
			free(new_str);
		} else {
			fprintf(cf, "%s\n", str);
		}
		fflush(cf);
	} else {
		/* actually this default should never be used. It's here just
		 * to satisfy the compiler */
		int priority = LOG_DEBUG;
		/* get log priority. XXX We cannot hold this as a static value
		 * because the user could configure something like this:
		 *
		 * <global>
		 *	syslogdebugpriority warn
		 * </global>
		 *
		 * <user anonymous>
		 *	syslogdebugpriority info
		 * </user>
		 *
		 * So we would have to reset it after each
		 * config_shrink_config(). Maybe we do this in the future but I
		 * don't want to do it now. We do the log_reinit() right now
		 * after rereading the configuration. But this is not enough.
		 * We don't have a function we call after each
		 * config_shrink_config.
		 * */
		const char* priority_str
			= config_get_option("syslogdebugpriority");

		if (priority_str) {
			priority = get_log_priority(priority_str);
		} else {
			if (debuglevel < 8) {
				priority = LOG_INFO;
			}
			if (debuglevel < 6) {
				priority = LOG_WARNING;
			}
			if (debuglevel < 4) {
				priority = LOG_ERR;
			}
		}

		vsnprintf(str, LOG_MAX_LINE_LENGTH - 1, fmt, args);
		/* doesn't look too nice:
		 * jftpgw[5949]: Write(10): 221 Goodbye...^M 
		 */
		replace_not_larger(str, "\r", "");
		syslog(priority, "%s", str);
	}
	va_end(args);
}

char* log_replace(int replace_type, union log_replace_un replace_val) {
	switch(replace_type) {
		char* s;

		case LOG_REPLACE_STRING:
			if (replace_val.replace_str) {
				s = strdup(replace_val.replace_str);
				enough_mem(s);
			} else {
				s = strdup("");
				enough_mem(s);
			}
			return s;
		case LOG_REPLACE_STRING_MALLOCED:
			if (replace_val.replace_str) {
				return (char*) replace_val.replace_str;
			} else {
				s = strdup("");
				enough_mem(s);
			}
			return s;
		case LOG_REPLACE_CHAR:
			s = (char*) malloc(2);
			enough_mem(s);
			s[0] = replace_val.replace_char;
			s[1] = '\0';
			return s;
		case LOG_REPLACE_INT:
			return mconv_int2char(replace_val.replace_int);
		case LOG_REPLACE_UINT:
			return mconv_uint2char(replace_val.replace_uint);
		case LOG_REPLACE_LINT:
			return mconv_lint2char(replace_val.replace_lint);
		case LOG_REPLACE_LUINT:
			return mconv_luint2char(replace_val.replace_luint);
		case LOG_REPLACE_FLOAT:
			return mconv_float2char(replace_val.replace_float);
	}
	return (char*) 0;
}

char* log_replace_line(const char* line, struct proxyinfo* pi) {

	char* replaced_line = (char*) 0;
	char* fragment;
	char* insert;
	int offset = 0;

	/* split up line */
	/* line:  blabla %u blubb %h bla */

	char*(*proto_log_replace_char)(const char, struct proxyinfo*);
	proto_log_replace_char = pi->protocol->proto_log_replace_char;

	if (line[0] == '%') {
		replaced_line = (proto_log_replace_char)(line[1], pi);
		offset = 2;
	}
	while ((fragment = quotstrtok(line, "%", &offset))) {
		if (!replaced_line) {
			replaced_line = fragment;
		} else {
			replaced_line = realloc(replaced_line,
				strlen(replaced_line) + strlen(fragment) + 1);
			enough_mem(replaced_line);
			strcat(replaced_line, fragment);
		}
		/* the key is at line[offset+1]; */
		if (line[offset]) {
			offset++;
			insert = (proto_log_replace_char)(line[offset], pi);
			if (!insert) {
				insert = strdup("<%x not found>");
				enough_mem(insert);
				/* replace `x' above */
				insert[2] = line[offset];
			}
			offset++;
			replaced_line = realloc(replaced_line,
				strlen(replaced_line) + strlen(insert) + 1);
			enough_mem(replaced_line);
			strcat(replaced_line, insert);
			free(insert);
		}
	}
	return replaced_line;
}



/* Idea by Bernd Eckenfels <ecki@lina.inka.de>
 * check of the command 'needle' is in the list of specified 
 * commands 'haystack'. Haystack may contain '*' which is a match-all
 * criteria */
int incommandpattern(const char *haystack, const char *needle) {
	char* negation = char_prepend(" ", needle);
	/* we now have "  NEEDLE " */
	negation[1] = '-';
	/* we now have " -NEEDLE " */

	if (strstr(haystack, negation)) {
		free(negation);
		return 0;
	}
	free(negation);

	if (strstr(haystack, " * ")) {
		return 1;
	}
	if (strstr(haystack, needle)) {
		return 1;
	}
	return 0;
}


void log_cmd(struct proxyinfo* pi) {
	/* go through the specifications of the logfiles and write
	 * them if they match */

	struct cmdlogtarget* files = loginfo.cmdlogfiles;
	struct cmdlogtarget* dirs  = loginfo.cmdlogdirs;

	while (files && files->logf_name) {
		pi->protocol->proto_log_cmd_to_target(files, pi);
		files = files->next;
	}

	while (dirs && dirs->logf_name) {
		pi->protocol->proto_log_cmd_to_target(dirs, pi);
		dirs = dirs->next;
	}
}

static
int log_name_exists(const struct cmdlogtarget* lt, const char* needle) {
	/* XXX we can still set a cmdlogfile and the debug logfile to the same
	 * file... */
	while (lt) {
		/* XXX we would need strcasecmp on WIN32... */
		if (strcmp(lt->logf_name, needle) == 0) {
			return 1;
		}
		lt = lt->next;
	}
	return 0;
}

static
int log_init_debuglevel() {
	int debuglevel;
	const char* dbl;

	if (!(dbl = config_get_option("debuglevel"))) {
		jlog(4, "No debug level specified. Using level 7");
		debuglevel = 7;
	} else {
		errno = 0;
		/* with conv_char2* we wouldn't be able to log a warning */
		debuglevel = strtol(dbl, (char**) 0, 10);
		if (errno || debuglevel < 0 || debuglevel > 9) {
			jlog(4, "Invalid debuglevel specified: \"%s\". Using "
				"level 7", dbl);
			debuglevel = 7;
		}
	}
	return debuglevel;
}

static
int get_log_priority(const char* log_priority_opt) {
	int log_priority;

	/* debug, info, notice, warning, err, crit, alert, emerg */
	
	if (strcasecmp(log_priority_opt, "debug") == 0) {
		log_priority = LOG_DEBUG;
	} else if (strcasecmp(log_priority_opt, "info") == 0) {
		log_priority = LOG_INFO;
	} else if (strcasecmp(log_priority_opt, "notice") == 0) {
		log_priority = LOG_NOTICE;
	} else if (strcasecmp(log_priority_opt, "warning") == 0) {
		log_priority = LOG_WARNING;
	} else if (strcasecmp(log_priority_opt, "err") == 0) {
		log_priority = LOG_ERR;
	} else if (strcasecmp(log_priority_opt, "crit") == 0) {
		log_priority = LOG_CRIT;
	} else if (strcasecmp(log_priority_opt, "alert") == 0) {
		log_priority = LOG_ALERT;
	} else if (strcasecmp(log_priority_opt, "emerg") == 0) {
		log_priority = LOG_EMERG;
	} else {
		jlog(4, "unknown log priority \"%s\", using warning as default",
					log_priority_opt);
		log_priority = LOG_WARNING;
	}
	return log_priority;
}

static
int get_log_facility(const char* log_facility_opt) {
	int log_facility;

	if (0) {
#ifdef HAVE_LOG_FACILITY_LOG_AUTH
	} else if (strcasecmp(log_facility_opt, "auth") == 0) {
		log_facility = LOG_AUTH;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_AUTHPRIV
	} else if (strcasecmp(log_facility_opt, "authpriv") == 0) {
		log_facility = LOG_AUTHPRIV;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_CRON
	} else if (strcasecmp(log_facility_opt, "cron") == 0) {
		log_facility = LOG_CRON;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_DAEMON
	} else if (strcasecmp(log_facility_opt, "daemon") == 0) {
		log_facility = LOG_DAEMON;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_FTP
	} else if (strcasecmp(log_facility_opt, "ftp") == 0) {
		log_facility = LOG_FTP;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_KERN
	} else if (strcasecmp(log_facility_opt, "kern") == 0) {
		log_facility = LOG_KERN;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL0
	} else if (strcasecmp(log_facility_opt, "local0") == 0) {
		log_facility = LOG_LOCAL0;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL1
	} else if (strcasecmp(log_facility_opt, "local1") == 0) {
		log_facility = LOG_LOCAL1;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL2
	} else if (strcasecmp(log_facility_opt, "local2") == 0) {
		log_facility = LOG_LOCAL2;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL3
	} else if (strcasecmp(log_facility_opt, "local3") == 0) {
		log_facility = LOG_LOCAL3;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL4
	} else if (strcasecmp(log_facility_opt, "local4") == 0) {
		log_facility = LOG_LOCAL4;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL5
	} else if (strcasecmp(log_facility_opt, "local5") == 0) {
		log_facility = LOG_LOCAL5;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL6
	} else if (strcasecmp(log_facility_opt, "local6") == 0) {
		log_facility = LOG_LOCAL6;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LOCAL7
	} else if (strcasecmp(log_facility_opt, "local7") == 0) {
		log_facility = LOG_LOCAL7;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_LPR
	} else if (strcasecmp(log_facility_opt, "lpr") == 0) {
		log_facility = LOG_LPR;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_MAIL
	} else if (strcasecmp(log_facility_opt, "mail") == 0) {
		log_facility = LOG_MAIL;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_NEWS
	} else if (strcasecmp(log_facility_opt, "news") == 0) {
		log_facility = LOG_NEWS;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_SYSLOG
	} else if (strcasecmp(log_facility_opt, "syslog") == 0) {
		log_facility = LOG_SYSLOG;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_USER
	} else if (strcasecmp(log_facility_opt, "user") == 0) {
		log_facility = LOG_USER;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_UUCP
	} else if (strcasecmp(log_facility_opt, "uucp") == 0) {
		log_facility = LOG_UUCP;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_CONSOLE
	} else if (strcasecmp(log_facility_opt, "console") == 0) {
		log_facility = LOG_CONSOLE;
#endif
#ifdef HAVE_LOG_FACILITY_LOG_SECURITY
	} else if (strcasecmp(log_facility_opt, "security") == 0) {
		log_facility = LOG_SECURITY;
#endif
	} else {
		/* facility not found, get default facility */
		if (loginfo.multithread) {
			log_facility = LOG_DAEMON;
			/* jlog will log to stdout if we have not yet opened a
			 * log file or syslog */
			jlog(5, "Facility \"%s\" not recognized, using default "
				"facility  %s", log_facility_opt, "LOG_DAEMON");
		} else {
			log_facility = LOG_USER;
			jlog(5, "Facility \"%s\" not recognized, using default "
				"facility  %s", log_facility_opt, "LOG_USER");
		}
	}
	return log_facility;
}

static
int parse_syslog_options(const char* specs, char** prefix, int* priority) {
	/* specs could be something like:
	 *
	 * syslog
	 * syslog:prefix
	 * syslog(priority)
	 * syslog(priority):prefix
	 * 
	 */

	const char* cmdlogpriority_str
			= config_get_option("syslogcmdlogpriority");
	int cmdlogpriority = get_log_priority(cmdlogpriority_str);

	if (strcmp(specs, "syslog") == 0) {
		*prefix = '\0';
		*priority = cmdlogpriority;
		return 0;
	}
	if (checkbegin(specs, "syslog:")) {
		/* syslog:prefix */
		*prefix = strdup(specs + strlen("syslog:"));
		*priority = cmdlogpriority;
		return 0;
	}
	if (checkbegin(specs, "syslog(")) {
		/* syslog(priority)
		 * or
		 * syslog(priority):prefix */
		char* brk;
		char* priority_opt;
		int size;
		if (strchr(specs, ':')) {
			/* syslog(priority):prefix */
			*prefix = strchr(specs, ':') + 1;
		} else {
			/* syslog(priority) */
			*prefix = '\0';
		}
		brk = strchr(specs, ')');
		size = brk - (specs + strlen("syslog(")) + 1;
		priority_opt = (char*) malloc(size);
		snprintf(priority_opt, size, "%s", specs + strlen("syslog("));
		*priority = get_log_priority(priority_opt);
		free(priority_opt);
		return 0;
	}
	return -1;
}

static
int log_init_syslog(struct loginfo_st* cfg) {
	const char* log_facility_opt = config_get_option("syslogfacility");
	int log_facility;

	cfg->syslog_facility = strdup(log_facility_opt);
	enough_mem(cfg->syslog_facility);

	log_facility = get_log_facility(log_facility_opt);

	openlog(loginfo.binaryname, LOG_PID, log_facility);
	syslog(LOG_INFO, "jftpgw v"PACKAGE_VERSION" opened syslog");
	cfg->syslog_opened = 1;
	return 0;
}

static
int log_init_logfile(struct loginfo_st* cfg) {
	const char* option;

	if (cfg->logf) {
		/* already logging */
		return 0;
	}
	cfg->syslog_debug = 0;
	/* read the logfile name from the configuration file */
	option = config_get_option("logfile");
	if (!option) {
		jlog(3, "no logfile set in the configuration file");
		return -1;
	}

	/* now option is set in every case */
	cfg->logf_name = strdup(strip_chrooted_path(option));
	enough_mem(cfg->logf_name);

	if (!(cfg->logf = open_logfile(cfg->logf_name))) {
		return -1;
	}
	jlog(7, "jftpgw v"PACKAGE_VERSION" opened the logfile");
	return 0;
}


static
int log_init_cmdlog(struct loginfo_st* cfg, int open) {
	char* option, *specs, *style;
	struct cmdlogtarget* files = cfg->cmdlogfiles;
	struct cmdlogtarget* new;
	char* logfile;
	int i;

	struct slist_t* opt_names =config_get_option_array("cmdlogfile");
	struct slist_t* opt_styles=config_get_option_array("cmdlogfile-style");
	struct slist_t* opt_specs =config_get_option_array("cmdlogfile-specs");

	int count_names = slist_count(opt_names);
	int count_styles = slist_count(opt_styles);
	int count_specs = slist_count(opt_specs);

	if (!(count_names == count_styles && count_styles == count_specs)) {
		jlog(4, "Counted %d times \"cmdlogfile\" as option",
				count_names);
		jlog(4, "Counted %d times \"cmdlogfile-style\" as option",
				count_styles);
		jlog(4, "Counted %d times \"cmdlogfile-specs\" as option",
				count_specs);
		jlog(4, "This is not balanced, please fix your configuration");
		return -1;
	}

	/* reverse them */
	opt_names  = slist_reverse(opt_names);
	opt_styles = slist_reverse(opt_styles);
	opt_specs  = slist_reverse(opt_specs);

	for (i = 0; i < count_names; i++) {
		option = slist_pop(opt_names);
		specs  = slist_pop(opt_specs);
		style  = slist_pop(opt_styles);
		logfile = strdup(strip_chrooted_path(option));
		enough_mem(logfile);
		if (log_name_exists(cfg->cmdlogfiles, logfile)) {
			jlog(6, "Command log file \"%s\" defined multiple "
					"times. Opening only once.",
					logfile);

			free(logfile);
			free(option);
			free(specs);
			free(style);
			continue;
		}
		if (log_name_exists(cfg->cmdlogfiles, option)) {
			jlog(6, "Command log file \"%s\" defined multiple "
					"times. Opening only once.",
					option);

			free(logfile);
			free(option);
			free(specs);
			free(style);
			continue;
		}

		free(option);

		new = malloc(sizeof (struct cmdlogtarget));
		enough_mem(new);
		if (cfg->cmdlogfiles) {
			/* there are existing entries */
			files->next = new;
		} else {
			/* this is the first entry */
			cfg->cmdlogfiles = new;
		}
		/* files always keeps the current entry and iterates on */
		files = new;
		files->logf_name = logfile;
		files->specs = char_enclose(" ", specs, " "); free(specs);
		files->style = style;
		files->next = (struct cmdlogtarget*) 0;
		files->syslog.used = 0;
		files->syslog.priority = 0;
		files->syslog.prefix = (char*) 0;
		files->logf = (FILE*) 0;
		if (open && checkbegin(logfile, "syslog")) {
			files->syslog.used = 1;
			if (parse_syslog_options(logfile,
					&(files->syslog.prefix),
					&(files->syslog.priority)) < 0) {
				return -1;
			}
			loginfo.syslog_cmdlog = 1;
			if (!loginfo.syslog_opened) {
				if (log_init_syslog(&loginfo) < 0) {
					return -1;
				}
			}
		} else if (open
			&& (files->logf = open_logfile(files->logf_name))
								== NULL) {
			return -1;
		}
		/* the malloc()ed memory in files will be freed by
		 * reset_loginfo() */
	}
	slist_destroy_free(opt_names);
	slist_destroy_free(opt_styles);
	slist_destroy_free(opt_specs);
	return 0;
}


#define FILENAME_EXTEND		64
static
int log_init_dirlog_open(struct cmdlogtarget* dir) {
	time_t now;
	char* filename, *fname;
	size_t size_max;
	char* fname_pattern;

	ASSUME(dir != NULL);
	fname_pattern = dir->logf_name;
	size_max = strlen(fname_pattern) + FILENAME_EXTEND;
	fname = (char*) malloc(size_max);
	filename = (char*) malloc(size_max);
	enough_mem(fname);
	enough_mem(filename);

	now = time(NULL);
	strftime(fname, size_max, fname_pattern, localtime(&now));
	snprintf(filename, size_max, fname, getpid());
	free(fname); fname = (char*) 0;
	free(dir->logf_name);
	dir->logf_name = filename;

	if ((dir->logf = open_logfile(dir->logf_name))) {
		return 0;
	} else {
		return -1;
	}
}

static
int log_init_dirlog(struct loginfo_st* cfg, struct proxyinfo* pi, int open) {
	char *specs, *style;
	char* prefix, *suffix;
	char* logfile, *logf_name;
	char* logdir;
	struct cmdlogtarget* dirs = cfg->cmdlogdirs;
	struct cmdlogtarget* new;
	int i;
	int chrooted = pi->si->chrooted;

	struct slist_t* opt_names
		= config_get_option_array("connectionlogdir");
	struct slist_t* opt_styles
		= config_get_option_array("connectionlogdir-style");
	struct slist_t* opt_specs
		= config_get_option_array("connectionlogdir-specs");
	struct slist_t* opt_prefix
		= config_get_option_array("connectionlogdir-fileprefix");
	struct slist_t* opt_suffix
		= config_get_option_array("connectionlogdir-filesuffix");

	int count_names = slist_count(opt_names);
	int count_styles = slist_count(opt_styles);
	int count_specs = slist_count(opt_specs);
	int count_prefix = slist_count(opt_prefix);
	int count_suffix = slist_count(opt_suffix);

	if (! (count_names == count_styles
			&& count_styles == count_specs
			&& count_specs == count_prefix
			&& count_prefix == count_suffix)) {

		jlog(4, "Counted %d times \"connectionlogdir\" as option",
				count_names);
		jlog(4, "Counted %d times \"connectionlogdir-style\" as option",
				count_styles);
		jlog(4, "Counted %d times \"connectionlogdir-specs\" as option",
				count_specs);
		jlog(4, "Counted %d times \"connectionlogdir-fileprefix\""
				" as option", count_prefix);
		jlog(4, "Counted %d times \"connectionlogdir-filesuffix\""
				" as option", count_suffix);
		jlog(4, "This is not balanced, please fix your configuration");
		return -1;
	}

	/* reverse them */
	opt_names  = slist_reverse(opt_names);
	opt_styles = slist_reverse(opt_styles);
	opt_specs  = slist_reverse(opt_specs);
	opt_prefix = slist_reverse(opt_prefix);
	opt_suffix = slist_reverse(opt_suffix);

	for (i = 0; i < count_names; i++) {
		logdir = slist_pop(opt_names);
		specs  = slist_pop(opt_specs);
		style  = slist_pop(opt_styles);
		prefix = slist_pop(opt_prefix);
		suffix = slist_pop(opt_suffix);

		replace_not_larger(prefix, "%", "-");
		replace_not_larger(suffix, "%", "-");
		/* check FILENAME_EXTEND if you change this pattern
		 * */
		logfile = char_enclose(
				prefix, "%Y-%m-%d--%H:%M:%S-%%d", suffix);

		logf_name = char_enclose(logdir, "/", logfile);
		free(logfile);

		free(logdir);
		free(prefix);
		free(suffix);

		if (chrooted) {
			char *logf_name_chroot
				= strdup(strip_chrooted_path(logf_name));
			enough_mem(logf_name_chroot);
			free(logf_name);
			logf_name = logf_name_chroot;
		}

		if (log_name_exists(cfg->cmdlogdirs, logf_name)) {
			jlog(6, "Command log dir \"%s\" defined multiple "
					"times. Opening only once.",
					logf_name);
			free(logf_name);
			free(style);
			free(specs);
			continue;
		}

		new = malloc(sizeof (struct cmdlogtarget));
		enough_mem(new);
		if (dirs) {
			/* there are existing entries */
			while (dirs->next) {
				dirs = dirs->next;
			}
			dirs->next = new;
		} else {
			/* this is the first entry */
			cfg->cmdlogdirs = new;
		}
		/* dirs always keeps the current entry and iterates on */
		dirs = new;
		dirs->next = (struct cmdlogtarget*) 0;
		dirs->specs = char_enclose(" ", specs, " ");
		dirs->logf_name = logf_name;
		dirs->style = style;
		free(specs);

		/* these don't make sense for dir logging */
		dirs->syslog.used = 0;
		dirs->syslog.prefix = (char*) 0;
		dirs->syslog.priority = 0;

		if (open && log_init_dirlog_open(dirs) < 0) {
			return -1;
		}
	}
	/* don't free if we're already at the end */
	slist_destroy_free(opt_names);
	slist_destroy_free(opt_styles);
	slist_destroy_free(opt_specs);
	slist_destroy_free(opt_prefix);
	slist_destroy_free(opt_suffix);
	return 0;
}

static
void log_deinit_cmdlogtarget(struct cmdlogtarget* cfg) {
	if (!cfg) {
		return;
	} else {
		log_deinit_cmdlogtarget(cfg->next);
	}
	/* XXX logf_name can be null */
	jlog(8, "Closing logfile %s", cfg->logf_name);
	if (cfg->logf) {
		fclose(cfg->logf);
	}
	free(cfg->specs);
	free(cfg->style);
	free(cfg->logf_name);
	free(cfg);
}

static
void log_deinit_cmdlog(struct cmdlogtarget* cfg) {
	log_deinit_cmdlogtarget(cfg); 
}

static
void log_deinit_dirlog(struct cmdlogtarget* cfg) {
	log_deinit_cmdlogtarget(cfg); 
}

int log_init(struct proxyinfo* pi) {
	loginfo.debuglevel = log_init_debuglevel();
	loginfo.servertype = pi->si->servertype;
	loginfo.multithread = pi->si->multithread;
	loginfo.binaryname = pi->si->binaryname;

	/* open the general logfile or syslog */
	if (config_compare_option("logstyle", "syslog")) {
		if (log_init_syslog(&loginfo) < 0) {
			return -1;
		}
		loginfo.syslog_debug = 1;
	} else {
		int changed_id = 0;
		if (getuid() == 0 && geteuid() != 0) {
			/* the log file entry will not show up in the log file
			 * of course because it is closed at the moment */
			if (changeid(PRIV, UID, "Changing UID to root for "
				"re-opening the log file", pi) < 0) {
				return -1;
			}
			changed_id = 1;
		}
		if (log_init_logfile(&loginfo) < 0) {
			return -1;
		} 
		if (changed_id) {
			changeid(UNPRIV, EUID,
				"After re-opening the logfile", pi);
		}
	}
	return 0;
}

int log_init_connect(struct proxyinfo* pi) {

	/* Open the command log logfile(s) */
	if (log_init_cmdlog(&loginfo, 1) < 0) {
		return -1;
	}

	/* Do the same with directories */
	if (log_init_dirlog(&loginfo, pi, 1) < 0) {
		return -1;
	}
	return 0;
}

int log_reinit(struct proxyinfo* pi) {
	jlog(6, "Re-opening logfile...");
	log_deinit(pi);
	return log_init(pi);
}

void log_deinit(struct proxyinfo* pi) {
	/* close the logfiles and delete the structures */
	if (loginfo.cmdlogdirs) {
		log_deinit_dirlog(loginfo.cmdlogdirs);
		loginfo.cmdlogdirs = (struct cmdlogtarget*) 0;
	}
	if (loginfo.cmdlogfiles) {
		log_deinit_cmdlog(loginfo.cmdlogfiles);
		loginfo.cmdlogfiles = (struct cmdlogtarget*) 0;
	}
	if (loginfo.logf_name) {
		jlog(8, "Closing log file %s", loginfo.logf_name);
		free_and_zero_char(&loginfo.logf_name);
	}
	if (loginfo.logf) {
		fclose(loginfo.logf);
		loginfo.logf = (FILE*) 0;
	}
	if (loginfo.syslog_opened) {
		jlog(8, "Closing syslog");
		closelog();
		loginfo.syslog_opened = 0;
	}
}

void log_stderr_closed(void) {
	loginfo.stderr_open = 0;
}

FILE* open_logfile(const char* fname) {
	FILE* logf;
	int err;

	umask(0077);
	jlog(8, "Opening log file %s", fname);
	logf = fopen(fname, "a");
	umask(0000);
	err = errno;
	if (!logf) {
		jlog(1, "Couldn't open the log file %s: %s", fname,
				strerror(errno));
	}
	errno = err;
	return logf;
}

